/*
 *	A simple DMA memory driver for the Linux kernel.
 */
 
 
#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/malloc.h>

#include <linux/dmaram.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>

#define MAX_DMARAM	32

#define DMA_MAJOR	125


struct dmaram
{
	int buffsize;
	void *raw_buf;
	unsigned long raw_buf_phys;
};

struct dmaram *dma_dev[MAX_DMARAM];

static int dmaram_create(struct dmaram *dp)
{
	char *start_addr, *end_addr;
	int size, sz, i;
	if(dp->buffsize < PAGE_SIZE)
		dp->buffsize = PAGE_SIZE;
	dp->raw_buf = NULL;
	start_addr = NULL;

	while (start_addr == NULL && dp->buffsize > PAGE_SIZE) {
		for (sz = 0, size = PAGE_SIZE; size < dp->buffsize; sz++, size <<= 1);
		{
			dp->buffsize = PAGE_SIZE * (1 << sz);
			start_addr = (char *) __get_free_pages(GFP_KERNEL, sz);
			if (start_addr == NULL)
				dp->buffsize /= 2;
		}
	}
	if (start_addr == NULL) {
		return -ENOMEM;
	} else {
		/* make some checks */
		end_addr = start_addr + dp->buffsize - 1;
	}
	dp->raw_buf = start_addr;
	dp->raw_buf_phys = virt_to_bus(start_addr);

	for (i = MAP_NR(start_addr); i < MAP_NR(end_addr); i++)
		set_bit(PG_reserved, &mem_map[i].flags);;
	return 0;
}

static void dmaram_destroy(struct dmaram *dp)
{
	int sz, size, i;
	unsigned long start_addr, end_addr;

	if (dp->raw_buf == NULL)
		return;
	for (sz = 0, size = PAGE_SIZE; size < dp->buffsize; sz++, size <<= 1);

	start_addr = (unsigned long) dp->raw_buf;
	end_addr = start_addr + dp->buffsize;

	for (i = MAP_NR(start_addr); i < MAP_NR(end_addr); i++)
		clear_bit(PG_reserved, &mem_map[i].flags);;

	free_pages((unsigned long) dp->raw_buf, sz);
	dp->raw_buf = NULL;
}


static int dma_open(struct inode *inode, struct file *file)
{
	void *ram;
	int minor = MINOR(inode->i_rdev);
	if(minor >= MAX_DMARAM)
		return -ENXIO;
		
	/*
	 *	The allocate may sleep - avoid races
	 */
	 
	ram = (struct dmaram *)kmalloc(sizeof(struct dmaram), GFP_KERNEL);
	if(ram==NULL)
		return -ENOMEM;
		
	if(dma_dev[minor] != NULL)
	{
		kfree(ram);
		return -EBUSY;	
	}
	
	dma_dev[minor]=ram;
	MOD_INC_USE_COUNT;
	memset(dma_dev[minor], 0, sizeof(struct dmaram));
	return 0;
}


static int dma_release(struct inode *inode, struct file *file)
{
	int minor = MINOR(inode->i_rdev);
	struct dmaram *dp = dma_dev[minor];
	if(dp->raw_buf)
		dmaram_destroy(dp);	
	kfree(dma_dev[minor]);
	dma_dev[minor]=NULL;
	MOD_DEC_USE_COUNT;
	return 0;
}

static long long dma_lseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

static int dma_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	struct dmaram *dp = dma_dev[MINOR(inode->i_rdev)];
	switch(cmd)
	{
		case DMAIOCSET:
		{
			long size;
			if(dp->raw_buf)
				return -EBUSY;
			if(get_user(size, (long *)arg))
				return -EFAULT;
			if(size < PAGE_SIZE || size%PAGE_SIZE)
				return -EINVAL;
			if(size > 32 * PAGE_SIZE)
				size = 32 *PAGE_SIZE;
				
			dp->buffsize = size;
			
			return dmaram_create(dp);
		}
		
		case DMAIOCGET:
		{
			struct dma_info dmi;
			dmi.size = dp->buffsize;
			dmi.phys = dp->raw_buf_phys;
			if(copy_to_user((void *)arg, &dmi, sizeof(dmi)))
				return -EFAULT;
			return 0;
		}
	}
	return -ENOTTY;
}


static int dma_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct dmaram *dp = dma_dev[MINOR(file->f_dentry->d_inode->i_rdev)];
	int size;
	
	if(dp->raw_buf == NULL)
		return -EIO;
	size = vma->vm_end - vma->vm_start;
	if(size != dp->buffsize)
		return -EINVAL;
	if(remap_page_range(vma->vm_start, dp->raw_buf_phys, 
		size, vma->vm_page_prot))
		return -EAGAIN;
	return 0;	
}

static ssize_t dma_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	return -EINVAL;
}

static ssize_t dma_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	return -EINVAL;
}

static struct file_operations dma_fops=
{
	dma_lseek,
	dma_read,
	dma_write,
	NULL,	/* readdir */
	NULL,
	dma_ioctl,
	dma_mmap,
	dma_open,
	NULL,		/* flush */
	dma_release
};

/*
 *	Initialise DMAram for Linux
 */
 
int dmaram_init(void)
{
	printk(KERN_INFO "Linux DMA memory interface v0.02 (c) 1999 Red Hat Software\n");
	if(register_chrdev(DMA_MAJOR,"dmaram", &dma_fops))
	{
		printk("dma_dev: unable to get major %d\n", DMA_MAJOR);
		return -EIO;
	}
	return 0;
}

#ifdef MODULE		
int init_module(void)
{
	return dmaram_init();
}

void cleanup_module(void)
{
	unregister_chrdev(DMA_MAJOR, "dmaram");
}

#endif
