#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/pci.h>

#include <asm/io.h>

#include "maestro.h"
#include "maestro_tables.h"

#include "sound_config.h"
#include "soundmodule.h"
#include "sb_mixer.h"
#include "sb.h"


#ifndef PCI_VENDOR_ESS
#define PCI_VENDOR_ESS			0x125D
#define PCI_DEVICE_ID_ESS_ESS1969	0x1969
#define PCI_DEVICE_ID_ESS_ESS1948	0x1948		/* Solo ?? 	*/
#define PCI_DEVICE_ID_ESS_ESS1968	0x1968		/* Maestro 2	*/
#define PCI_DEVICE_ID_ESS_ESS1978      	0x1978		/* Maestro 2E	*/
#endif

#define ESS_SBBASE	0x220

struct address_info maestro_cfg;


/*
 *	ESS Maestro AC97 codec programming interface.
 */
	 
static void maestro_ac97_set(int io, u8 cmd, u16 val)
{
	int i;
	/*
	 *	Wait for the codec bus to be free 
	 */
	 
	for(i=0;i<10000;i++)
	{
		if(!(inb(io+ESS_AC97_INDEX)&1))
			break;
	}
	/*
	 *	Write the bus
	 */ 
	outw(val, io+ESS_AC97_DATA);
	udelay(1);
	outb(cmd, io+ESS_AC97_INDEX);
	udelay(1);
}

static u16 maestro_ac97_get(int io, u8 cmd)
{
	int sanity=100000;
	u16 data;
	int i;
	
	/*
	 *	Wait for the codec bus to be free 
	 */
	 
	for(i=0;i<10000;i++)
	{
		if(!(inb(io+ESS_AC97_INDEX)&1))
			break;
	}

	outb(cmd|0x80, io+ESS_AC97_INDEX);
	udelay(1);
	
	while(inb(io+ESS_AC97_INDEX)&1)
	{
		sanity--;
		if(!sanity)
		{
			printk(KERN_ERR "ess_pci: ac97 codec timeout.\n");
			return 0;
		}
	}
	data=inw(io+ESS_AC97_DATA);
	udelay(1);
	return data;
}

/*
 *	The Maestro can be wired to a standard AC97 compliant codec
 *	(see www.intel.com for the pdf's on this), or to a PT101 codec
 *	which appears to be the ES1918 (data sheet on the esstech.com.tw site)
 *
 *	The PT101 setup is untested.
 */
 
static u16 maestro_ac97_init(int iobase)
{
	maestro_ac97_set(iobase, 0x02, 0x0000);
	maestro_ac97_set(iobase, 0x04, 0x0000);
	maestro_ac97_set(iobase, 0x06, 0x0000);
	maestro_ac97_set(iobase, 0x08, 0x9F1F);
	maestro_ac97_set(iobase, 0x0A, 0x9F1F);
	maestro_ac97_set(iobase, 0x0C, 0x9F1F);
	maestro_ac97_set(iobase, 0x0E, 0x9F1F);
	maestro_ac97_set(iobase, 0x10, 0x9F1F);
	maestro_ac97_set(iobase, 0x12, 0x9F1F);
	maestro_ac97_set(iobase, 0x14, 0x9F1F);
	maestro_ac97_set(iobase, 0x16, 0x9F1F);
	maestro_ac97_set(iobase, 0x18, 0x0808);
	maestro_ac97_set(iobase, 0x1A, 0x0000);
	maestro_ac97_set(iobase, 0x1C, 0x0404);
	maestro_ac97_set(iobase, 0x1E, 0x0404);
	maestro_ac97_set(iobase, 0x20, 0x0000);
	maestro_ac97_set(iobase, 0x26, 0x000F);
	return 0;
}

static u16 maestro_pt101_init(int iobase)
{
	maestro_ac97_set(iobase, 0x2A, 0x0001);
	maestro_ac97_set(iobase, 0x2C, 0x0000);
	maestro_ac97_set(iobase, 0x2C, 0xFFFF);
	maestro_ac97_set(iobase, 0x10, 0x9F1F);
	maestro_ac97_set(iobase, 0x12, 0x0808);
	maestro_ac97_set(iobase, 0x14, 0x9F1F);
	maestro_ac97_set(iobase, 0x16, 0x9F1F);
	maestro_ac97_set(iobase, 0x18, 0x0404);
	maestro_ac97_set(iobase, 0x1A, 0x0000);
	maestro_ac97_set(iobase, 0x1C, 0x0000);
	maestro_ac97_set(iobase, 0x02, 0x0404);
	maestro_ac97_set(iobase, 0x04, 0x0808);
	maestro_ac97_set(iobase, 0x0C, 0x801F);
	maestro_ac97_set(iobase, 0x0E, 0x801F);
	return 0;
}

static void maestro_ac97_reset(int ioaddr)
{
	outw(0x2000,  ioaddr+0x36);
	udelay(100);
	outw(0x0000,  ioaddr+0x36);
	udelay(100);
}


/*
 *	Indirect register access. Not all registers are readable so we
 *	need to keep register state ourselves
 */
 
static u16 maestro_map[32];
#define WRITEABLE_MAP	0xEFFFFF
#define READABLE_MAP	0x64003F

/*
 *	The ESS engineers were a little indirection happy. These indirected
 *	registers themselves include indirect registers at another layer
 */
 
static void maestro_write(int ioaddr, u16 reg, u16 data)
{
	unsigned long flags;
	save_flags(flags);
	cli();
	outw(reg, ioaddr+0x02);
	outw(data, ioaddr+0x00);
	maestro_map[reg]=data;
	restore_flags(flags);
}

static u16 maestro_read(int ioaddr, u16 reg)
{
	if(READABLE_MAP & (1<<reg))
	{
		unsigned long flags;
		save_flags(flags);
		cli();
		outw(reg, ioaddr+0x02);
		maestro_map[reg]=inw(ioaddr+0x00);
		restore_flags(flags);
	}
	return maestro_map[reg];
}

/*
 *	These routines handle accessing the second level indirections to the
 *	wave ram.
 */

/*
 *	The register names are the ones ESS uses (see 104T31.ZIP)
 */
 
#define IDR0_DATA_PORT		0x00
#define IDR1_CRAM_POINTER	0x01
#define IDR2_CRAM_DATA		0x02
#define IDR3_WAVE_DATA		0x03
#define IDR4_WAVE_PTR_LOW	0x04
#define IDR5_WAVE_PTR_HI	0x05
#define IDR6_TIMER_CTRL		0x06
#define IDR7_WAVE_ROMRAM	0x07

static void apu_index_set(int ioaddr, u16 index)
{
	int i;
	maestro_write(ioaddr, IDR1_CRAM_POINTER, index);
	for(i=0;i<1000;i++)
		if(maestro_read(ioaddr, IDR1_CRAM_POINTER)==index)
			return;
	printk(KERN_WARNING "ess_pci: APU register select failed.\n");
}

static void apu_data_set(int ioaddr, u16 data)
{
	int i;
	for(i=0;i<1000;i++)
	{
		if(maestro_read(ioaddr, IDR0_DATA_PORT)==data)
			return;
		maestro_write(ioaddr, IDR0_DATA_PORT, data);
	}
}

/*
 *	This is the public interface for APU manipulation. It handles the
 *	interlock to avoid two APU writes in parallel etc. Don't diddle
 *	directly with the stuff above.
 */

static void apu_set_register(int ioaddr, u8 channel, u8 reg, u16 data)
{
	unsigned long flags;
	
	reg|=(channel<<4);
	
	save_flags(flags);
	cli();
	apu_index_set(ioaddr, reg);
	apu_data_set(ioaddr, data);
	restore_flags(flags);
}

static u16 apu_get_register(int ioaddr, u8 channel, u8 reg)
{
	unsigned long flags;
	u16 v;
	
	reg|=(channel<<4);
	
	save_flags(flags);
	cli();
	apu_index_set(ioaddr, reg);
	v=maestro_read(ioaddr, IDR0_DATA_PORT);
	restore_flags(flags);
	return v;
}

/*
 *	The ASSP is fortunately not double indexed
 */
 
static void assp_set_register(int ioaddr, u32 reg, u32 value)
{
	unsigned long flags;
	
	save_flags(flags);
	cli();
	outl(reg, ioaddr+0x80);
	outl(value, ioaddr+0x84);
	restore_flags(flags);
}

static u32 assp_get_register(int ioaddr, u32 reg)
{
	unsigned long flags;
	u32 value;
	
	save_flags(flags);
	cli();
	outl(reg, ioaddr+0x80);
	value=inl(ioaddr+0x84);
	restore_flags(flags);
	
	return value;
}

/*
 *	The wavecache is also sane
 */
 
static void wave_set_register(int ioaddr, u16 reg, u16 value)
{
	unsigned long flags;
	
	save_flags(flags);
	cli();
	outw(reg, ioaddr+0x10);
	outw(value, ioaddr+0x12);
	restore_flags(flags);
}

static u16 wave_get_register(int ioaddr, u16 reg)
{
	unsigned long flags;
	u16 value;
	
	save_flags(flags);
	cli();
	outw(reg, ioaddr+0x10);
	value=inw(ioaddr+0x12);
	restore_flags(flags);
	
	return value;
}

/*
 *	Additional Reset Functions
 */

static void maestro_reset(int ioaddr)
{
	outw(0x8000, 0x18+ioaddr);
	udelay(10);
	outw(0x0000, 0x18+ioaddr);
	udelay(10);
}

static void sound_reset(int ioaddr)
{
	outw(0x2000, 0x18+ioaddr);
	udelay(10);
	outw(0x0000, 0x18+ioaddr);
	udelay(10);
}

static void asp_load(int ioaddr, u16 l, u16 h, u16 *data, int len)
{
	int i;
	outw(l, ioaddr+0x80);
	outw(h, ioaddr+0x82);
	for(i=0;i<len;i++)
		outw(*data++, ioaddr+0x84);
}

static void asp_memset(int ioaddr, u16 l, u16 h, u16 v, int len)
{
	int i;
	outw(l, ioaddr+0x80);
	outw(h, ioaddr+0x82);
	for(i=0;i<len;i++)
		outw(v, ioaddr+0x84);
}


/*
 *	Load the table stuff. This isn't needed on a Maestro-1 except
 *	for FM emulation (which I haven't put in yet) but might be on
 *	the 2 and 2E
 */

 
static void load_tables(int iobase)
{
	asp_load(iobase, 0x0, 0x2, asp_block_0, sizeof(asp_block_0)/2);
	asp_load(iobase, 0x0800, 0x2, asp_block_1, sizeof(asp_block_1)/2);
	asp_memset(iobase, 0x1000, 0x03, 0, 1024);
	asp_memset(iobase, 0x11BC, 0x03, 0x18, 36);
	asp_load(iobase, 0x13DC, 0x03, asp_block_2, sizeof(asp_block_2)/2);
	asp_load(iobase, 0x1300, 0x03, asp_block_3, sizeof(asp_block_3)/2);
}


/*
 *	Native play back driver 
 */
 
 
static void ess_play_setup(int io, int low_apu, int mode, u32 rate, void *buffer, int size)
{
	u32 pa;
	int high_apu = low_apu;
	int channel;
		
	if(mode&1)
		high_apu++;

	size >>=1;
		
	for(channel=low_apu; channel <= high_apu; channel++)
	{
		int i;
		
		pa = virt_to_bus(buffer);
		wave_set_register(io, 0x01FC+channel, 0x0000);
		wave_set_register(io,0x0000, pa&0xFFFF);
		pa>>=1;
		if(mode&1)			/* Enable stereo */
			pa|=0x00800000;
		pa|=0x00400000;			/* System RAM */
		
		/* Begin loading the APU */
		
		for(i=0;i<15;i++)
			apu_set_register(io, channel, i, 0x0000);
			
		/* Load the frequency, turn on 6dB, turn off the effects */
		apu_set_register(io, channel, 2, (rate&0xFF)<<8|0x10);
		apu_set_register(io, channel, 3, rate>>8);
		
		/* Load the buffer into the wave engine */
		apu_set_register(io, channel, 4, ((pa>>16)&0xFF)<<8);
		apu_set_register(io, channel, 5, pa&0xFFFF);
		apu_set_register(io, channel, 6, (pa+size)&0xFFFF);
		apu_set_register(io, channel, 7, size);
		
		apu_set_register(io, channel, 8, 0x0000);
		apu_set_register(io, channel, 9, 0xD000);
		
		if(mode&1)
		{
			/* Set left or right */
			if(channel == low_apu)
				apu_set_register(io, channel, 10, 0x8F00);
			else			
				apu_set_register(io, channel, 10, 0x8F10);
		}
		else
			apu_set_register(io, channel, 10, 0x8F08);
		apu_set_register(io, channel, 11, 0x0000);
		apu_set_register(io, channel, 0, 0x400F);
	}
	
//	outw(1, io+0x04)
//	outw(inw(io+0x18)|4, io+0x18);

	for(channel=low_apu; channel<=high_apu; channel++)
	{
		/* Turn on the DMA */
		if(mode&2)	/* 16 bit */
			apu_set_register(io, channel, 0,
				(apu_get_register(io, channel, 0)&0xFF0F)|0x20);
		else
			apu_set_register(io, channel, 0,
				(apu_get_register(io, channel, 0)&0xFF0F)|0x40);
	}
}

static void ess_play_finish(int io, int channel, int mode)
{
	apu_set_register(io, channel, 0,
		apu_get_register(io, channel, 0)&0xFF0F);
	if(mode&1)
	{
		channel++;
		apu_set_register(io, channel, 0,
			apu_get_register(io, channel, 0)&0xFF0F);
	}
}

static void ess_play_test(int io)
{
	u8 *buf=kmalloc(8192, GFP_KERNEL|GFP_DMA);
	int ct;
	long time=jiffies;
	for(ct=0; ct<8192;ct++)
	{
		buf[ct]=ct^39;
	}
	ess_play_setup(io, 10, 0,  0x39CD, buf, 8192);
	
	while((jiffies-time)<5*HZ)
	{
		schedule();
	}
	printk("\n");
	ess_play_finish(io, 10, 0);
	kfree(buf);
}


/*
 *	Configure up legacy mode on the ESS-Tech Maestro and Maestro-2
 *
 *	This is also pretty much the same set up to configure the native
 *	mode which is what we will finally use. This is the "2.2 at least
 *	you get sound" stopgap for the moment.
 */
 
static int maestro_install(struct pci_dev *pcidev, char *name)
{
	u16 w;
	u32 n;
	int iobase;
	int i;
		
	iobase = pcidev->base_address[0] & PCI_BASE_ADDRESS_IO_MASK;

	/*
	 *	Ok card ready. Begin setup proper
	 */
	 
	printk(KERN_INFO "ess_pci: Configuring %s at 0x%04X\n", name, iobase);

	if(check_region(ESS_SBBASE, 16))
	{
		printk(KERN_WARNING "ess_pci: skipping - I/O 0x%03X is not free.\n",
			ESS_SBBASE);
		return 0;
	}
	
	/*
	 *	Disable ACPI
	 */
	 
	pci_write_config_dword(pcidev, 0x54, 0x00000000);
	
	/*
	 *	Use TDMA for now. TDMA works on all boards, so while its
	 *	not the most efficient its the simplest.
	 */ 
	 
	pci_read_config_word(pcidev, 0x50, &w);

	/* Clear DMA bits */
	w&=~(1<<10|1<<9|1<<8);

	/* TDMA on */
	w|=(1<<8);

	/*
	 *	MPU at 330
	 */
	
	w&=~((1<<4)|(1<<3));
	
	/*
	 *	SB at 0x220
	 */

	w&=~(1<<2);
	
	/*
	 *	Reserved write as 0
	 */
	 
	w&=~(1<<1);
	
	/*
	 *	Some of these are undocumented bits
	 */
	 
	w&=~(1<<13)|(1<<14);		/* PIC Snoop mode bits */
	w&=~(1<<11);			/* Safeguard off */
	w|= (1<<7);			/* Posted write */
	w|= (1<<6);			/* ISA timing on */
	w&=~(1<<1);			/* Subtractive decode off */
	w&=~(1<<5);			/* Don't swap left/right */
	
	pci_write_config_word(pcidev, 0x50, w);
	
	pci_read_config_word(pcidev, 0x52, &w);
	w&=~(1<<15);		/* Turn off internal clock multiplier */
	w&=~(1<<14);		/* External clock */
	
	w&=~(1<<7);		/* HWV off */
	w&=~(1<<6);		/* Debounce off */
	w&=~(1<<5);		/* GPIO 4:5 */
	w&=~(1<<4);		/* Disconnect from the CHI */
	w&=~(1<<3);		/* IDMA off (undocumented) */
	w&=~(1<<2);		/* MIDI fix off (undoc) */
	w&=~(1<<0);		/* IRQ to ISA off (undoc) */
	pci_write_config_word(pcidev, 0x52, w);

	/*
	 *	DDMA off
	 */

	pci_read_config_word(pcidev, 0x60, &w);
	w&=~(1<<0);
	pci_write_config_word(pcidev, 0x60, w);
	
	/*
	 *	Legacy mode
	 */

	pci_read_config_word(pcidev, 0x40, &w);
	w&=~(1<<15);	/* legacy decode on */
	w&=~(1<<14);	/* Disable SIRQ */
	w&=~(1<<13|1<<12|1<<11);

	/*
	 *	We don't use the emulation IRQ's
	 */

	w|=(1<<11);			/* MPU IRQ 7 */
	w&=~(1<<10|1<<9|1<<8);		/* SB IRQ 5 */
	w|=(1<<3)|(1<<1)|(1<<0);	/* SB on , FM on, MPU on */
	w&=~(1<<7|1<<6);
	w|=(1<<6);			/* DMA 1 */

	/*
	 *	Default to DMA 1 
	 */
	 
	pci_write_config_word(pcidev, 0x40, w);

	sound_reset(iobase);
	

	/*
	 *	Reset the CODEC
	 */
	 
	maestro_ac97_reset(iobase);
		
	/*
	 *	Ring Bus Setup
	 */
	 
	n=inl(iobase+0x34);
	n&=~0xF000;
	n|=12<<12;		/* Direct Sound, Stereo */

	n=inl(iobase+0x34);
	n&=~0x0F00;		/* Modem off */
	outl(n, iobase+0x34);

	n=inl(iobase+0x34);
	n&=~0x00F0;
	n|=9<<4;		/* DAC, Stereo */
	outl(n, iobase+0x34);
	
	n=inl(iobase+0x34);
	n&=~0x000F;		/* ASSP off */
	outl(n, iobase+0x34);
	
	
	n=inl(iobase+0x34);
	n|=(1<<29);		/* Enable ring bus */
	outl(n, iobase+0x34);
	
	
	n=inl(iobase+0x34);
	n|=(1<<28);		/* Enable serial bus */
	outl(n, iobase+0x34);
	
	n=inl(iobase+0x34);
	n&=~0x00F00000;		/* MIC off */
	outl(n, iobase+0x34);
	
	n=inl(iobase+0x34);
	n&=~0x000F0000;		/* I2S off */
	outl(n, iobase+0x34);
	
	w=inw(iobase+0x18);
	w&=~(1<<7);		/* ClkRun off */
	outw(w, iobase+0x18);

	w=inw(iobase+0x18);
	w&=~(1<<6);		/* Harpo off */
	outw(w, iobase+0x18);
	
	w=inw(iobase+0x18);
	w&=~(1<<4);		/* ASSP irq off */
	outw(w, iobase+0x18);
	
	w=inw(iobase+0x18);
	w&=~(1<<3);		/* ISDN irq off */
	outw(w, iobase+0x18);
	
	w=inw(iobase+0x18);
	w|=(1<<2);		/* Direct Sound IRQ on */
	outw(w, iobase+0x18);

	w=inw(iobase+0x18);
	w&=~(1<<1);		/* MPU401 IRQ off */
	outw(w, iobase+0x18);

	w=inw(iobase+0x18);
	w|=(1<<0);		/* SB IRQ on */
	outw(w, iobase+0x18);


	outb(0, iobase+0xA4);
	outb(3, iobase+0xA2);
	outb(0, iobase+0xA6);
	
	for(i=0;i<16;i++)
	{
		/* Write 0 into the buffer area 0x1E0->1EF */
		outw(0x01E0+i, 0x10+iobase);
		outw(0x0000, 0x12+iobase);
	}

	wave_set_register(iobase, IDR7_WAVE_ROMRAM, 
		(wave_get_register(iobase, IDR7_WAVE_ROMRAM)&0xFF00));
	wave_set_register(iobase, IDR7_WAVE_ROMRAM,
		wave_get_register(iobase, IDR7_WAVE_ROMRAM)|0x100);
	wave_set_register(iobase, IDR7_WAVE_ROMRAM,
		wave_get_register(iobase, IDR7_WAVE_ROMRAM)&~0x200);
	wave_set_register(iobase, IDR7_WAVE_ROMRAM,
		wave_get_register(iobase, IDR7_WAVE_ROMRAM)|~0x400);
	
	maestro_write(iobase, IDR2_CRAM_DATA, 0x0000);
	maestro_write(iobase, 0x08, 0xB004);
	/* Now back to the DirectSound stuff */
	maestro_write(iobase, 0x09, 0x001B);
	maestro_write(iobase, 0x0A, 0x8000);
	maestro_write(iobase, 0x0B, 0x3F37);
	maestro_write(iobase, 0x0C, 0x0098);
	
	maestro_write(iobase, 0x0C, 
		(maestro_read(iobase, 0x0C)&~0xF000)|0x8000);
	maestro_write(iobase, 0x0C, 
		(maestro_read(iobase, 0x0C)&~0x0F00)|0x0500);

	maestro_write(iobase, 0x0D, 0x7632);
			
	/* Wave cache control on - test off, sg off, 
		enable, enable extra chans 1Mb */
		
	outw(inw(0x14+iobase)|(1<<8),0x14+iobase);
	outw(inw(0x14+iobase)&0xFE03,0x14+iobase);
	outw((inw(0x14+iobase)&0xFFFC), 0x14+iobase);
	outw(inw(0x14+iobase)|(1<<7),0x14+iobase);

	outw(0xA1A0, 0x14+iobase);

	if(maestro_ac97_get(iobase, 0x00)==0x0080)
	{
		printk(KERN_INFO "ess_pci: PT101 Codec detected\n");
		maestro_pt101_init(iobase);
	}
	else
	{
		printk(KERN_INFO "ess_pci: AC97 Codec detected\n");
		maestro_ac97_init(iobase);	
	}

	/* Now clear the channel data */
	
	for(i=0;i<64;i++)
	{
		for(w=0;w<0x0E;w++)
			apu_set_register(iobase, i, w, 0);
	}
	
	/* Now we program up the APU's  */
	
	/*
	 *	APU 6 is the mixer 
	 */
	 
	apu_set_register(iobase, 6, 0, 
		apu_get_register(iobase, 6, 0)&~(1<<14));
	apu_set_register(iobase, 6, 0, 
		apu_get_register(iobase, 6, 0)&~(1<<13));
	apu_set_register(iobase, 6, 0, 
		apu_get_register(iobase, 6, 0)&~(1<<12));
	apu_set_register(iobase, 6, 0, 
		apu_get_register(iobase, 6, 0)&0xFCFF);
	apu_set_register(iobase, 6, 0, 
		apu_get_register(iobase, 6, 0)&0xF3FF);
	apu_set_register(iobase, 6, 0, 
		(apu_get_register(iobase, 6, 0)&0xFFF0)|0x0F);
	apu_set_register(iobase, 6, 2,
		apu_get_register(iobase, 6, 2)&0x00FF);
	apu_set_register(iobase, 6, 3, 0x0000);
	apu_set_register(iobase, 6, 2,
		apu_get_register(iobase, 6, 2)&0xFF3F);
	apu_set_register(iobase, 6, 2,
		apu_get_register(iobase, 6, 2)|(1<<5));
	apu_set_register(iobase, 6, 2,
		apu_get_register(iobase, 6, 2)&~(1<<4));
	apu_set_register(iobase, 6, 2,
		apu_get_register(iobase, 6, 2)&0xFFF0);
	apu_set_register(iobase, 6, 4, 0x4B5F);
	apu_set_register(iobase, 6, 5, 0xFDD5);
	apu_set_register(iobase, 6, 6, 0xFE00);
	apu_set_register(iobase, 6, 7, 0x0100);
	apu_set_register(iobase, 6, 8, 0x78FF);
	apu_set_register(iobase, 6, 9, 0xFFFF);
	apu_set_register(iobase, 6, 0xA,0x8A10);
	apu_set_register(iobase, 6, 0xB,0x8009);
	
	/*
	 *	APU 62 - Right channel DSP
	 */
		
	apu_set_register(iobase, 62, 0, 
		apu_get_register(iobase, 62, 0)&~(1<<14));
	apu_set_register(iobase, 62, 0, 
		apu_get_register(iobase, 62, 0)&~(1<<13));
	apu_set_register(iobase, 62, 0, 
		apu_get_register(iobase, 62, 0)&~(1<<12));
	apu_set_register(iobase, 62, 0, 
		apu_get_register(iobase, 62, 0)&0xFCFF);
	apu_set_register(iobase, 62, 0, 
		apu_get_register(iobase, 62, 0)&0xF3FF);
	apu_set_register(iobase, 62, 0, 
		(apu_get_register(iobase, 62, 0)&0xFFF0)|0x0F);
	apu_set_register(iobase, 62, 2,
		apu_get_register(iobase, 62, 2)&0x00FF);
	apu_set_register(iobase, 62, 3, 0x1000);
	apu_set_register(iobase, 62, 2,
		apu_get_register(iobase, 62, 2)&0xFF3F);
	apu_set_register(iobase, 62, 2,
		apu_get_register(iobase, 62, 2)&~(1<<5));
	apu_set_register(iobase, 62, 2,
		apu_get_register(iobase, 62, 2)&~(1<<4));
	apu_set_register(iobase, 62, 2,
		apu_get_register(iobase, 62, 2)&0xFFF0);
	apu_set_register(iobase, 62, 4, 0x7F00);
	apu_set_register(iobase, 62, 5, 0x8000);
	apu_set_register(iobase, 62, 6, 0x8080);
	apu_set_register(iobase, 62, 7, 0x0080);
	apu_set_register(iobase, 62, 8, 0x7F00);
	apu_set_register(iobase, 62, 9, 0xFFFF);
	apu_set_register(iobase, 62, 0xA,0x7F00);
	apu_set_register(iobase, 62, 0xB,0x00F0);

	/*
	 *	APU 63 - Left channel DSP
	 */
		
	apu_set_register(iobase, 63, 0, 
		apu_get_register(iobase, 63, 0)&~(1<<14));
	apu_set_register(iobase, 63, 0, 
		apu_get_register(iobase, 63, 0)&~(1<<13));
	apu_set_register(iobase, 63, 0, 
		apu_get_register(iobase, 63, 0)&~(1<<12));
	apu_set_register(iobase, 63, 0, 
		apu_get_register(iobase, 63, 0)&0xFCFF);
	apu_set_register(iobase, 63, 0, 
		apu_get_register(iobase, 63, 0)&0xF3FF);
	apu_set_register(iobase, 63, 0, 
		(apu_get_register(iobase, 63, 0)&0xFFF0)|0x0F);
	apu_set_register(iobase, 63, 2,
		apu_get_register(iobase, 63, 2)&0x00FF);
	apu_set_register(iobase, 63, 3, 0x1000);
	apu_set_register(iobase, 63, 2,
		apu_get_register(iobase, 63, 2)&0xFF3F);
	apu_set_register(iobase, 63, 2,
		apu_get_register(iobase, 63, 2)&~(1<<5));
	apu_set_register(iobase, 63, 2,
		apu_get_register(iobase, 63, 2)&~(1<<4));
	apu_set_register(iobase, 63, 2,
		apu_get_register(iobase, 63, 2)&0xFFF0);
	apu_set_register(iobase, 63, 4, 0x7F00);
	apu_set_register(iobase, 63, 5, 0x0000);
	apu_set_register(iobase, 63, 6, 0x0080);
	apu_set_register(iobase, 63, 7, 0x0080);
	apu_set_register(iobase, 63, 8, 0x7F00);
	apu_set_register(iobase, 63, 9, 0xFFFF);
	apu_set_register(iobase, 63, 0xA,0x7F10);
	apu_set_register(iobase, 63, 0xB,0x00F0);

	apu_set_register(iobase, 6, 0,
		(apu_get_register(iobase,6, 0)&0xFF0F)|(9<<4));
	
	udelay(30);
	
	apu_set_register(iobase, 62, 0,
		(apu_get_register(iobase,6, 0)&0xFF0F)|(1<<4));
	apu_set_register(iobase, 63, 0,
		(apu_get_register(iobase,6, 0)&0xFF0F)|(1<<4));

        load_tables(iobase);

	outb(0x41, ESS_SETUP_A4+iobase);
	udelay(1);

	printk(KERN_INFO "ess_pci: Programmed %s at 0x%X to legacy mode.\n",
		name, iobase);

	ess_play_test(iobase);
	maestro_cfg.dma = 1;
	maestro_cfg.irq = pcidev->irq;
	maestro_cfg.io_base = ESS_SBBASE;
	maestro_cfg.dma2 = -1;
	if(sb_dsp_detect(&maestro_cfg, SB_PCI_ESSMAESTRO, iobase)==-1)
	{
		printk(KERN_WARNING "esspci: Sound blaster emulation not responding.\n");
		return 0;
	}
	attach_sb_card(&maestro_cfg);
	return 1;
}

	
	
/*
 *	For now this does one card only. It is written this way as it will change
 */
 
int init_ess19xx(void)
{
	struct pci_dev *pcidev=NULL;
	int count=0;
		
	if(!pci_present())
		return -ENODEV;
	
	pcidev = NULL;

	/*
	 *	Find the ESS Maestro 2.
	 */

	while((pcidev = pci_find_device(PCI_VENDOR_ESS, PCI_DEVICE_ID_ESS_ESS1968, pcidev))!=NULL)
	{
		count+=maestro_install(pcidev, "ESS Maestro 2");
		if(count)
			return 0;
	}

	/*
	 *	Find the ESS Maestro 2E
	 */

	while((pcidev = pci_find_device(PCI_VENDOR_ESS, PCI_DEVICE_ID_ESS_ESS1978, pcidev))!=NULL)
	{
//		count+=maestro_install(pcidev, "ESS Maestro 2E");
		printk(KERN_INFO "ESS Maestro 2E not yet supported.\n");
		if(count)
			return 0;
	}

	/*
	 *	ESS Maestro 1
	 */

	while((pcidev = pci_find_device(0x1285, 0x100, pcidev))!=NULL)
	{
		count+=maestro_install(pcidev, "ESS Maestro");
		if(count)
			return 0;
	}
	if(count==0)
		return -ENODEV;
	return 0;
}


int init_module(void)
{
	if(init_ess19xx()<0)
	{
		printk(KERN_ERR "No ESS Maestro cards found.\n");
		return -ENODEV;
	}
	return 0;
}

void cleanup_module(void)
{
	if(maestro_cfg.slots[0] != -1)
		unload_sb(&maestro_cfg);
}

