#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/pci.h>

#include <asm/io.h>

#include "sound_config.h"
#include "soundmodule.h"
#include "sb_mixer.h"
#include "sb.h"

/*
 *	Legacy initialisation tables
 */
 
struct pci_legacy
{
	int vendor;
	int device;
	char *description;
	int (*install)(struct pci_dev *, struct pci_legacy *);
	int flags;
};

#define MAX_CARDS	8

static struct address_info pci_cards[MAX_CARDS];
static int card_count=0;

static struct address_info pci_mpu_cards[MAX_CARDS];
static int mpu_count=0;

static int find_space(int *array, int len)
{
	int i;
	for(i=0;array[i]!=0;i++)
	{
		if(check_region(array[i],len)==0)
			return i;
	}
	return -1;
}

static int find_dma(int *array)
{
	int i;
	for(i=0;array[i]!= -1;i++)
	{
		if(request_dma(array[i],"pci_legacy")==0)
		{
			free_dma(array[i]);
			return i;
		}
	}
	return -1;
}

/*
 *	Setup the Yamaha cards. For now don't do the MPU401/OPL3 ports.
 *	Does anyone actually use the glorified doorbell mode on an SB ?
 *
 *	For now we only support PC/PCI. If you don't have a PC/PCI sideband
 *	connector on your board bad luck - you can write the DDMA support.
 */
 
static int ymf_init(struct pci_dev *pci_dev, struct pci_legacy *pl)
{
	u16 r;
	int slot;
	int dma;
	int doorbell;
	int mpu;
	static int ymf_sb[]={0x220, 0x240, 0x260, 0x280, 0};
	static int ymf_opl3[]={0x3A8,0x3A0, 0x398, 0x388, 0};
	static int ymf_dma[]={1,3,0, -1};
	static int ymf_mpu[]={0x330, 0x300, 0x332, 0x334, 0};
	struct address_info *card;
	
	if((slot=find_space(ymf_sb, 16))==-1)
	{
		printk(KERN_WARNING "%s: no free soundblaster addresses.\n",
			pl->description);
		return 0;
	}
	if((dma=find_dma(ymf_dma))==-1)
	{
		printk(KERN_WARNING "%s: no free DMA channels.\n",
			pl->description);
		return 0;
	}
		
	doorbell = find_space(ymf_opl3, 4);
	mpu = find_space(ymf_mpu, 4);
	
	printk(KERN_INFO "Found a %s. Due to lack of documentation this card will\n",
		pl->description);
	printk(KERN_INFO "be configured as an 8bit soundblaster.\n");
	switch(pl->flags)
	{
		case 0:
		case 1:
			pci_read_config_word(pci_dev, 0x42, &r);
			/* Mask the fields we will set */
			r&=~0x783F;
			/* Set the I/O */
			r|=(slot<<2);
			/* Serialized IRQ crap off */
			r|=(1<<15);
			/* Set the OPL3 */
			if(doorbell != -1)
			{
				printk(KERN_INFO "Enabling OPL3 at 0x%03X.\n",
					ymf_opl3[doorbell]);
				r|=(3-doorbell);
			}
			if(mpu != -1)
			{
				r|=(mpu<<4);
				printk(KERN_INFO "Enabling MPU at I/O 0x%03X.\n",
					ymf_mpu[mpu]);
			}
			pci_write_config_word(pci_dev, 0x42, r);
			pci_read_config_word(pci_dev, 0x40, &r);
			/* Set DMA */
			r&=~0x00C0;
			r|=(ymf_dma[dma])<<6;
			/* SB port only */
			r&=~0xF;
			r|=1;
			if(doorbell != -1)
				r|=2;
			if(mpu != -1)
				r|=24;
			/* SIRQ off, I/O on */
			r&=~(1<<14);
			r&=~(1<<15);
			pci_write_config_word(pci_dev, 0x40, r);
			

			pci_read_config_word(pci_dev, 0x4C, &r);
			r&=~1;
			pci_write_config_word(pci_dev, 0x4C, r);

			pci_read_config_word(pci_dev, 0x48, &r);
			r=1;
			pci_write_config_word(pci_dev, 0x48, r);
			

			pci_read_config_word(pci_dev, 0x48, &r);
			udelay(2000);
			r=0;
			pci_write_config_word(pci_dev, 0x48, r);
			

			pci_read_config_word(pci_dev, 0x48, &r);
			r=3;
			udelay(2000);
			pci_write_config_word(pci_dev, 0x48, r);

			printk(KERN_INFO "%s: set to I/O 0x%03X DMA %d.\n",
				pl->description,
				ymf_sb[slot], ymf_dma[dma]);
			break;
	}
	
	card=&pci_cards[card_count];
	
	card->dma = ymf_dma[dma];
	card->irq = pci_dev->irq;
	card->io_base = ymf_sb[slot];
	card->dma2 = -1;
	
	if(sb_dsp_detect(card, SB_PCI_YAMAHA, 0)==-1)
	{
		printk(KERN_WARNING "pci_legacy: Sound blaster emulation not responding.\n");
		return 0;
	}
	
	attach_sb_card(card);
	
	if(mpu != -1)
	{
		card=&pci_mpu_cards[mpu_count];
		card->dma = -1;
		card->irq = -1;
		card->io_base = ymf_mpu[mpu];
		card->dma2 = -1;
		if(probe_sbmpu(card))
		{
			attach_sbmpu(card);
			mpu_count++;
		}
		else
			printk(KERN_WARNING "pci_legacy: MPU401 emulation not responding.\n");

	}
	
	card_count++;
	return 1;
}

/*
 *	Important; Put the less flexible cards first so that they get
 *	the SB emulation slots they need.
 */
 
struct pci_legacy pci_legacy_table[]={
	{ 0x1073, 0x0003, "Yamaha YFM-740C", ymf_init, 0 },
	{ 0x1073, 0x000D, "Yamaha YFM-724F", ymf_init, 1 },
	{ 0x1073, 0x0004, "Yamaha YFM-724", ymf_init, 1 },   
	{ 0, 0, NULL, 0, 0}
};


int init_legacy_pci(void)
{
	struct pci_dev *pcidev=NULL;
	int count=0;
	struct pci_legacy *pt=&pci_legacy_table[0];
		
	if(!pci_present())
		return -ENODEV;

	while(pt->vendor)
	{
		pcidev = NULL;

		printk("Scan for %04X %04X\n",
			pt->vendor, pt->device);
		while((pcidev = pci_find_device(pt->vendor, pt->device, pcidev))!=NULL)
		{
			printk("Trying to init %04X %04X\n",
				pt->vendor, pt->device);
			count+=pt->install(pcidev, pt);
			if(count >= MAX_CARDS)
				return 0;
		}
		pt++;
	}
	
	if(count==0)
		return -ENODEV;
		
	printk(KERN_INFO "pci_legacy: %d unsupported PCI card%s placed in legacy 8bit mode.\n",
		card_count,
		card_count==1?"s":"");
	return 0;
}


int init_module(void)
{
	if(init_legacy_pci()<0)
	{
		printk(KERN_ERR "No supported devices found.\n");
		return -ENODEV;
	}
	return 0;
}

void cleanup_module(void)
{
	int i;
	for(i=0;i<card_count;i++)
		unload_sb(&pci_cards[i]);
	for(i=0;i<mpu_count;i++)
		unload_sbmpu(&pci_mpu_cards[i]);
}


