/*
 * linux/fs/nfs/cluster.c
 *
 * Clustering of file data over NFS.
 *
 * The write-back code takes a two-level approach. At the lower level,
 * there is a request associated with each dirty page.
 * These pages are grouped in clusters in order to improve write scheduling.
 * The benefits are:
 *
 *  1.  If we write out clusters of pages simultaneously, the write-gathering
 *      code on the server side can optimize syncing the data (for NFSv2).
 *  2.  For NFSv3, we can schedule COMMIT calls more intelligently.
 *  3.  If the write block size is bigger than physical page size, we
 *      are able to group several pages into one write request.
 *
 * Clusters are linked into the inode so that the file sync code et. al.
 * can conveniently loop over all dirty pages. There is also a hash table
 * of clusters so that the write() code itself can locate the relevant
 * cluster fairly quick.
 *
 * Each cluster holds a list of requests (struct nfs_page) that represent
 * the dirty portion of a page. These can be considered as an equivalent
 * to the buffer heads currently used in the VFS. A request struct holds
 * a minimal set of information to keep memory overhead low; most of
 * the state is in fact kept in the cluster.
 *
 * For each NFS mount, there is a separate cache object that contains
 * a hash table of all clusters. With this cache, an async RPC task
 * (`clusterd') is associated, which wakes up occasionally to inspect
 * its list of dirty buffers.
 * (Note that RPC tasks aren't kernel threads. Take a look at the
 * rpciod code to understand what they are).
 *
 * Inside the cache object, we also maintain a count of the current number
 * of dirty pages, which may not exceed a certain threshold.
 * (FIXME: This threshold should be configurable).
 *
 * The code is streamlined for what I think is the prevalent case for
 * NFS traffic, which is sequential write access without concurrent
 * access by different processes.
 *
 * Copyright (C) 1996, 1997, Olaf Kirch <okir@monad.swb.de>
 *
 * Some parts
 * Copyright (C) 1999, Trond Myklebust <trond.myklebust@fys.uio.no>
 */

#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/pagemap.h>
#include <linux/file.h>

#include <linux/sched.h>

#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/clnt.h>
#include <linux/sunrpc/sched.h>

#include <linux/nfs.h>
#include <linux/nfs_fs.h>
#include <linux/nfs_fs_sb.h>
#include <linux/nfs_cluster.h>

/*
 * Various constants
 */
#define NFS_WRITEBACK_DELAY	(10 * HZ)
#define NFS_WRITEBACK_LOCKDELAY (120 * HZ)

#define NFSDBG_FACILITY         NFSDBG_PAGECACHE

/*
 * This is the per-mount writeback cache.
 */
struct nfs_reqlist {
	unsigned int            nr_requests;
	struct wait_queue *     request_wait;

	struct nfs_server *	server;
	/* The async RPC task that is responsible for scanning the
	 * clusters.
	 */
	struct rpc_task *       task;           /* cluster flush task */

	/* Authentication flavor handle for this NFS client */
	struct rpc_auth *       auth;

	/* The 'hash' table of all clusters.
	 * Technically, this needn't be a hash table; all we
	 * want is an even distribution of clusters across
	 * slots in this table.
	 */
	struct nfs_cluster *    clusters[CLUSTER_HASH_SIZE];

	/* Used by clusterd when scanning the hash table */
	unsigned int            hand;
};

/*
 * This is the wait queue all cluster daemons sleep on
 */
static struct rpc_wait_queue    cluster_queue = RPC_INIT_WAITQ("clusterd");

/*
 * Local function declarations.
 */
static void                     nfs_clusterd(struct rpc_task *);
static void                     nfs_clusterd_done(struct rpc_task *);


int
nfs_reqlist_init(struct nfs_server *server)
{
	struct nfs_reqlist	*cache;
	struct rpc_task		*task;

	dprintk("NFS: writecache_init\n");
	cache = (struct nfs_reqlist *) kmalloc(sizeof(*cache), GFP_KERNEL);
	if (cache == NULL)
		return -ENOMEM;
	memset(cache, 0, sizeof(*cache));

	/* Create the RPC task */
	task = rpc_new_task(server->client, nfs_clusterd_done, 0);
	if (task == NULL) {
		kfree(cache);
		return -ENOMEM;
	}
	task->tk_flags   |= RPC_TASK_ASYNC;
	task->tk_action   = nfs_clusterd;
	task->tk_calldata = cache;

	/* Put the task to sleep */
	task->tk_timeout = 60 * HZ;
	rpc_sleep_on(&cluster_queue, task, NULL, NULL);

	cache->task = task;
	cache->auth = server->client->cl_auth;
	server->rw_requests = cache;
	cache->server = server;
	return 0;
}

void
nfs_reqlist_exit(struct nfs_server *server)
{
	struct nfs_reqlist	*cache;
	struct rpc_task		*task;

	if ((cache = server->rw_requests) == 0)
		return;

	dprintk("NFS: reqlist_exit (ptr %p rpc %p)\n", cache, cache->task);

	server->rw_requests = NULL;
	task = cache->task;
	task->tk_exit = NULL;
	rpc_kill(task, 0);
}

static void
nfs_clusterd_done(struct rpc_task *task)
{
	struct nfs_reqlist	*cache = (struct nfs_reqlist *)task->tk_calldata;
	struct nfs_server	*server;

	if (cache) {
		server = cache->server;
		if (server)
			server->rw_requests = NULL;
		kfree(cache);
	}
}


/*
 * Check whether the file range we want to write to is locked by
 * us.
 */
static int
region_locked(struct inode *inode, struct nfs_page *req)
{
	struct file_lock	*fl;

	/* Don't optimize writes if we don't use NLM */
	if (NFS_SERVER(inode)->flags & NFS_MOUNT_NONLM)
		return 0;

	for (fl = inode->i_flock; fl; fl = fl->fl_next) {
		if (fl->fl_owner == current->files && (fl->fl_flags & FL_POSIX)
		    && fl->fl_type == F_WRLCK
		    && fl->fl_start <= req->start && req->end <= fl->fl_end) {
			return 1;
		}
	}

	return 0;
}

/*
 * Create a new cluster and associated RPC task.
 *
 * Clusters are created only in update_request, and we know the inode
 * is nfs-locked when we get there.
 */
static struct nfs_cluster *
create_cluster(struct file *file, nfs_flush_req flush,
	       __u64 rqstart, __u64 rqend, int *error)
{
	struct dentry		*dentry = file->f_dentry;
	struct inode		*inode = dentry->d_inode;
	struct nfs_reqlist	*cache = NFS_REQUESTLIST(inode);
	struct nfs_cluster	*cluster, *next;
	unsigned int		hash;

	dprintk("NFS:      create_cluster(%x/%ld, %ld)\n",
		inode->i_dev, inode->i_ino,
		(long) (rqstart & CLUSTER_MASK));

	/*
	 * Did somebody recently kill RPCIOD?
	 */
	if (!cache) {
		*error = nfs_reqlist_init(NFS_SERVER(inode));
		cache = NFS_REQUESTLIST(inode);
		if (!cache)
			return NULL;
	}
	/* Default error out of memory */
	*error = -ENOMEM;

	/* Should never fail */
	cluster = (struct nfs_cluster *) rpc_allocate(0, sizeof(*cluster));
	if (cluster == NULL)
		return NULL;

	dprintk("NFS:      allocated cluster %p\n", cluster);
	memset(cluster, 0, sizeof(*cluster));

	cluster->nextscan = jiffies + NFS_WRITEBACK_DELAY;
	cluster->file     = file;
	file->f_count++;
	cluster->flush    = flush;
	cluster->start    = rqstart;
	cluster->end      = rqend;

	/* Note that we don't have to keep a reference on the
	 * inode as write_inode takes care of any pending writes. */

	/* Attach cluster to the inode */
	cluster->count++;

	rpc_append_list(&NFS_CLUSTERS(inode), cluster);

	/* Insert the cluster into the hash table */
	hash = CLUSTER_HASH(inode->i_ino, rqstart);
	if ((next = cache->clusters[hash]) != NULL)
		next->hash_prev = cluster;
	cluster->hash_next  = next;
	cluster->hash_prev  = NULL;
	cache->clusters[hash] = cluster;

	*error = 0;
	return cluster;
}

/*
 * Delete an empty cluster
 */
void
delete_cluster(struct nfs_cluster *cluster)
{
	struct nfs_reqlist	*cache;
	struct nfs_cluster	*prev, *next;
	struct dentry		*dentry = cluster->file->f_dentry;
	struct inode		*inode = dentry->d_inode;
	unsigned int		hash;

	dprintk("NFS:      delete_cluster(%p)\n", cluster);

	if (cluster->count)
		return;

	if (cluster->pages)
		printk("NFS: Arrgh! Cluster not empty!\n");

	/* Remove from inode list of clusters */
	rpc_remove_list(&NFS_CLUSTERS(inode), cluster);

	/* Remove from cluster hash table */
	cache = NFS_REQUESTLIST(inode);
	hash = CLUSTER_HASH(inode->i_ino, cluster->start);
	prev = cluster->hash_prev;
	next = cluster->hash_next;
	if (prev != NULL)
		prev->hash_next = next;
	else
		cache->clusters[hash] = next;
	if (next != NULL)
		next->hash_prev = prev;

	/* Free cluster memory */
	fput(cluster->file);
	kfree(cluster);
}


/*
 * Insert a write request into a cluster
 */
static inline void
nfs_insert_request(struct nfs_cluster *cluster, struct nfs_page *req)
{
	struct inode		*inode = cluster->file->f_dentry->d_inode;
	cluster->count++;

	if (cluster->request[REQUEST_NR(req->start)])
		printk("NFS: Uh oh! Overwriting existing request!\n");

	cluster->request[REQUEST_NR(req->start)] = req;
	cluster->pages++;

	/* Update the cluster region */
	if (req->start < cluster->start)
		cluster->start = req->start;
	if (cluster->end < req->end)
		cluster->end   = req->end;

	NFS_REQUESTLIST(inode)->nr_requests++;
}

/*
 * Create a write request.
 * Page must be locked by the caller. This makes sure we never create
 * two different requests for the same page, and avoids possible deadlock
 * when we reach the hard limit on the number of dirty pages.
 */
struct nfs_page *
nfs_create_request(struct nfs_cluster *cluster, struct page *page,
		   __u64 rqstart, __u64 rqend, int *error)
{
	struct inode		*inode = cluster->file->f_dentry->d_inode;
	struct nfs_reqlist	*cache = NFS_REQUESTLIST(inode);
	struct nfs_page		*req = NULL;
	signed long		timeout;

	/* Deal with hard/soft limits.
	 * Increment the reference count on the cluster to make sure it
	 * doesn't go away while we sleep.
	 */
	cluster->count++;
	atomic_inc(&page->count);
	while (1) {
		*error = 0;

		/* If we're over the soft limit, wake up some requests */
		if (cache->nr_requests >= MAX_REQUEST_SOFT
		    && !NFS_CONGESTED(inode)) {
			dprintk("NFS:      hit soft limit (%d requests)\n",
				cache->nr_requests);
			rpc_wake_up_task(cache->task);
		}

		/* If we haven't reached the hard limit yet,
		 * try to allocate the request struct */
		if (cache->nr_requests < MAX_REQUEST_HARD) {
			req = (struct nfs_page *) kmalloc(sizeof(*req),
							  GFP_KERNEL);
			if (req != NULL)
				break;
		}

		/* We're over the hard limit. Wait for better times */
		dprintk("NFS:      create_request sleeping (total %d pid %d)\n",
			cache->nr_requests, current->pid);
		timeout = jiffies + (HZ >> 2);
		if (NFS_SERVER(inode)->flags & NFS_MOUNT_INTR) {
			interruptible_sleep_on_timeout(&cache->request_wait, timeout);
			*error = -ERESTARTSYS;
			if (signalled())
				break;
		} else
			sleep_on(&cache->request_wait);

		dprintk("NFS:      create_request waking up (tot %d pid %d)\n",
			cache->nr_requests, current->pid);
	}

	if (!req)
		goto done;

	/* Initialize the request struct. Initially, we assume a
	 * long write-back delay. This will be adjusted in
	 * update_nfs_request below if the region is not locked. */
	req->cluster = cluster;
	req->page    = page;
	req->task    = NULL;
	req->start   = rqstart;
	req->end     = rqend;
	req->timeout = jiffies + NFS_WRITEBACK_LOCKDELAY;
	req->flags   = 0;
	req->commit  = 0;

	/* Insert the request into the cluster */
	nfs_insert_request(cluster, req);

 done:
	release_cluster(cluster);
	if (!req)
		__free_page(page);
	return req;
}


/*
 * Try to update any existing write request, or create one if there is none.
 * In order to match, the request's credentials must match those of
 * the calling process.
 *
 * The inode is nfs-locked when we get here.
 */
int
nfs_update_request(struct file* file, struct page *page,
		   __u64 offset, __u64 bytes,
		   struct nfs_page **reqp,
		   nfs_flush_req flush)
{
	struct inode		*inode = file->f_dentry->d_inode;
        struct nfs_cluster	*cluster = NULL, *head, *c;
        struct nfs_page		*req;
        __u64			rqstart, rqend;
        __u64			base;
	unsigned int		rqnr;
	int			status;

	base    = page->offset & CLUSTER_MASK;
	rqstart = page->offset + offset;
	rqend   = rqstart + bytes;
	rqnr    = REQUEST_NR(rqstart);

	/* Loop over all clusters and see if we find
	 *  a.  A cluster with matching region and creds
	 *  b.  A request for the page we wish to update
	 */

	head = c = NFS_CLUSTERS(inode);
	if (!c)
		goto create;

	do {
		if ((c->start & CLUSTER_MASK) != base)
			continue;
		/* If the cluster matches our creds and pid, remember it.
		 * We must match against the pid so errors can be reported
		 * to the correct process.
		 * XXX do we really care?
		 */
		if (c->file == file && c->flush == flush)
			cluster = c;

		if ((req = c->request[rqnr]) != NULL) {
			/* We have a request for our page.
			 * If the cluster creds don't match, or the
			 * page addresses don't match (can that happen?),
			 * tell the caller to wait on the conflicting
			 * request.
			 */
			if (req->page != page || c != cluster ||
			    rqstart > req->end || rqend < req->start) {
				*reqp = req;
				return 0;
			}

			/* Okay, the request matches. Update the region */
			if (rqstart < req->start) {
				if (rqstart < cluster->start)
					cluster->start = rqstart;
				req->start = rqstart;
			}
			if (req->end < rqend) {
				if (cluster->end < rqend)
					cluster->end = rqend;
				req->end = rqend;
			}
			cluster->count++;

			goto doupdate;
		}
	} while ((c = CL_NEXT(c)) != head);

	if (cluster)
		cluster->count++;

 create:

	/* No matching request found.
	 * Create a cluster if we don't have one.
	 */
	if (cluster == NULL) {
		cluster = create_cluster(file, flush, rqstart, rqend, &status);
		if (cluster == NULL)
			goto out;
	}

	/* Create the request. It's safe to sleep in this call because
	 * we only get here if the page is locked.
	 */
	req = nfs_create_request(cluster, page, rqstart, rqend, &status);
	if (req == NULL)
		goto out;

 doupdate:
	status = 1;

	nfs_mark_request_dirty(cluster, req);

	/* If the region is not locked, adjust the timeout */
	if (!inode->i_flock || !region_locked(inode, req)) {
		unsigned long   timeout;

		timeout = jiffies + NFS_WRITEBACK_DELAY;
		if (time_before(timeout, req->timeout)) {
			if (time_before(timeout, cluster->nextscan))
				cluster->nextscan = timeout;
			req->timeout = timeout;
		}
	}
	*reqp = req;

out:
	release_cluster(cluster);
	return status;
}

/*
 * Wait for a request to complete.
 *
 * Interruptible by signals only if mounted with intr flag.
 */
int
nfs_wait_on_page(struct inode *inode, struct page *page)
{
	struct rpc_clnt		*clnt = NFS_CLIENT(inode);
	struct wait_queue	wait = { current, NULL };
	sigset_t		oldmask;
	int			retval, intr;

	intr = NFS_SERVER(inode)->flags & NFS_MOUNT_INTR;
	rpc_clnt_sigmask(clnt, &oldmask);
	add_wait_queue(&page->wait, &wait);
	atomic_inc(&page->count);
	for(;;) {
		current->state = (intr) ? TASK_INTERRUPTIBLE
					: TASK_UNINTERRUPTIBLE;

		retval = 0;
		if (!PageLocked(page))
			break;

		retval = -ERESTARTSYS;
		if (intr && signalled())
			break;

		schedule();
	}

	remove_wait_queue(&page->wait, &wait);
	current->state = TASK_RUNNING;
	rpc_clnt_sigunmask(clnt, &oldmask);
	__free_page(page);

	return retval;
}

int
nfs_wait_on_request(struct nfs_page *req)
{
	struct inode	*inode = req->cluster->file->f_dentry->d_inode;
	return nfs_wait_on_page(inode, req->page);
}

/*
 * Release all resources associated with a write request after it
 * has been committed to stable storage
 */
void
nfs_release_request(struct nfs_page *req)
{
	struct nfs_cluster	*cluster = req->cluster;
	struct inode		*inode = cluster->file->f_dentry->d_inode;
	struct nfs_reqlist	*cache = NFS_REQUESTLIST(inode);
	struct page		*page;

	cluster->request[REQUEST_NR(req->start)] = NULL;
	cluster->pages--;
	if ((page = req->page) != NULL) {
		if (req->flags & PG_INVALIDATE_AFTER)
			clear_bit(PG_uptodate, &page->flags);
		if (req->flags & PG_UNLOCK_AFTER)
			nfs_unlock_page(page);
		__free_page(page);
		req->page = NULL;
	}
        kfree(req);
	release_cluster(cluster);
	/* wake up anyone waiting to allocate a request */
	cache->nr_requests--;
	wake_up(&cache->request_wait);
}

/*
 * This is the strategy routine for NFS.
 * It is called by nfs_updatepage whenever the user wrote up to the end
 * of a page.
 *
 * We always try to submit a set of requests in parallel so that the
 * server's write code can gather writes. This is mainly for the benefit
 * of NFSv2.
 *
 * We never submit more requests than we think the remote can handle.
 * For UDP sockets, we make sure we don't exceed the congestion window;
 * for TCP, we limit the number of requests to 8.
 *
 * NFS_STRATEGY_PAGES gives the minimum number of requests for NFSv2 that
 * should be sent out in one go. This is for the benefit of NFSv2 servers
 * that perform write gathering.
 *
 * FIXME: Different servers may have different sweet spots.
 * Record the average congestion window in server struct?
 */
#define NFS_STRATEGY_PAGES      4
int
nfs_strategy(struct nfs_cluster *cluster)
{
	struct inode    *inode = cluster->file->f_dentry->d_inode;
	unsigned int    dirty, wpages;

	/* If we've reached the end of the cluster, flush everything */
	if ((cluster->end & ~CLUSTER_MASK) == 0) {
		nfs_sync_file(inode, cluster->file, cluster->start,
			      cluster->end, FLUSH_WAIT);
		return 0;
        }

	dirty  = cluster->dirty - cluster->pending;
	wpages = NFS_SERVER(inode)->wsize / PAGE_SIZE;
	if (dirty < NFS_STRATEGY_PAGES * wpages)
		return 0;

	return sync_cluster(cluster, 0);
}

/*
 * Flush out all dirty pages. Called for
 *
 *  -   fsync, fclose, and when releasing a write lock.
 *      How is FLUSH_WAIT in this case.
 *  -   nfs_invalidate_pages.
 *      How is FLUSH_INVALIDATE|FLUSH_STABLE in this case.
 *
 * The start/end arguments specify a file range to flush. To make the
 * logic simpler, we flush the entire cluster even if it overlaps the
 * file range only partially.
 *
 * The inode must be nfs-locked when we get here, in order to avoid
 * that new write requests get created while we're busy flushing out
 * the old ones.
 */
int
nfs_sync_file(struct inode *inode, struct file *file, __u64 start, __u64 end, int how)
{
	struct nfs_cluster	*cluster, *last, *head, *waiton;
	int                     error = 0, wait, count;

	wait = how & FLUSH_WAIT;
	how &= ~FLUSH_WAIT;

	if (file && ! inode)
		inode = file->f_dentry->d_inode;

	/* Do this repeatedly. We may write all pages to the server only
	 * to find out that it has rebooted when we get around to the
	 * commit call.
	 */
again:
	head = cluster = NFS_CLUSTERS(inode);
	waiton = last = 0;
	count = 0;

	if (cluster == NULL)
		goto wait;

	head->count++;

	do {
		if (cluster->pages && (file == 0 || file == cluster->file)
		    && cluster->start < end && start < cluster->end) {

			/* Cluster overlaps file region.
			 * Flush it out and remember pointer. */
			cluster->count++;

			release_cluster(last);
			last = cluster;

			if (wait && NFS_CONGESTED(inode))
				error = nfs_wait_on_congest(inode);
			if (error < 0)
				break;

			error = sync_cluster(cluster, how);
			if (error < 0)
				break;

			if (wait && cluster->pending) {
				cluster->count++;
				release_cluster(waiton);
				waiton = cluster;
			}
		}
	} while ((cluster = CL_NEXT(cluster)) && cluster != head);

wait:
	release_cluster(head);

	if(error < 0)
		goto out;

	/* Wait for the last cluster to complete */
	if (waiton) {
		error = wait_cluster(waiton);
		if (error < 0)
			goto out;
		release_cluster(waiton);
		release_cluster(last);
		goto again;
	}

out:
	release_cluster(waiton);
	release_cluster(last);
	return error;
}


/*
 * Send all dirty pages of a cluster to the server and lock them.
 *
 * When invalidating pages we must make sure every page gets locked
 * down. Therefore we send even written but uncommitted pages to stable
 * storage right away.
 */
int
sync_cluster(struct nfs_cluster *cluster, int how)
{
	struct nfs_page *req;
	struct inode	*inode;
	unsigned int    nr, flushed = 0;
	int             error = 0;

	cluster->count++;

	inode = cluster->file->f_dentry->d_inode;
	for (nr = 0; nr < CLUSTER_PAGES; nr++) {
		if (!(req = cluster->request[nr]))
			continue;

		if (how & FLUSH_INVALIDATE)
			req->flags |= PG_INVALIDATE_AFTER;
		else if (!IS_DIRTY(req))
			continue;

		if (NFS_CONGESTED(inode))
			continue;

		if (nfs_lock_page(req)) {
			error = cluster->flush(req, how);
			if (error < 0)
				goto out;

			flushed++;
		}
	}

out:
	release_cluster(cluster);

	if (error >= 0)
		error = flushed;
	return error;
}

/*
 * Scan cluster for dirty pages and send as many of them to the
 * server as possible.
 */
int
scan_cluster(struct nfs_cluster *cluster, unsigned int max)
{
	struct inode	*inode = cluster->file->f_dentry->d_inode;
	struct nfs_page *req;
	unsigned int    nr, flushed = 0;
	int             error = 0;

	cluster->count++;
	for (nr = 0; nr < CLUSTER_PAGES; nr++) {
		if (!(req = cluster->request[nr]))
			continue;

		if (time_before(req->timeout, cluster->nextscan))
			cluster->nextscan = req->timeout;
		if (time_after(req->timeout, jiffies))
			continue;
		if (NFS_CONGESTED(inode))
			continue;
		if (IS_DIRTY(req) && nfs_lock_page(req)) {
			error = cluster->flush(req, FLUSH_AGING);
			if (error < 0)
				break;
			if (max && ++flushed >= max)
				break;
		}
	}

	release_cluster(cluster);

	if (error >= 0)
		error = flushed;
	return error;
}

/*
 * Wait for all write requests to complete, and commit them when
 * they're done.
 */
int
wait_cluster(struct nfs_cluster *cluster)
{
	struct dentry		*dentry = cluster->file->f_dentry;
	struct inode		*inode = dentry->d_inode;
	struct rpc_clnt		*clnt = NFS_CLIENT(inode);
	struct wait_queue	wait = {current, NULL};
	sigset_t		oldmask;
	int			intr, error = 0;

	cluster->count++;

	intr = NFS_SERVER(inode)->flags & NFS_MOUNT_INTR;
	rpc_clnt_sigmask(clnt, &oldmask);
	add_wait_queue(&cluster->wait, &wait);

	for(;;) {
		current->state = (intr) ? TASK_INTERRUPTIBLE
					: TASK_UNINTERRUPTIBLE;

		error = 0;
		if (!cluster->pending)
			break;

		error = -ERESTARTSYS;
		if (intr && signalled())
			break;

		schedule();
	}

	remove_wait_queue(&cluster->wait, &wait);
	current->state = TASK_RUNNING;
	rpc_clnt_sigunmask(clnt, &oldmask);

	release_cluster(cluster);

        return error;
}

/*
 * This is the cluster scanner.
 *  -   If there are pages due to be written, send them out now.
 *  -   If the total number of dirty pages has exceeded the soft
 *      limit, sync the cluster now.
 *  -   If the user has written past the end of the cluster, we
 *      also flush the remaining dirty pages and initiate a commit
 *      call.
 */
static void
nfs_clusterd(struct rpc_task *task)
{
	struct nfs_reqlist	*cache;
	struct nfs_cluster	*cluster, *previous;
	struct inode		*inode;
	unsigned long		delay = 60 * HZ;
	unsigned int		hand;
	int			error = 0;

	dprintk("NFS: %4d clusterd starting\n", task->tk_pid);
	cache = (struct nfs_reqlist *) task->tk_calldata;

	hand = cache->hand; 
	do {
		cluster  = cache->clusters[hand];
		previous = 0;
		while (cluster && error >= 0) {
			cluster->count++;

			if (!cluster->pages)
				goto release;
			if (cluster->pending == cluster->dirty)
				goto release;

			inode = cluster->file->f_dentry->d_inode;
			mark_inode_dirty(inode);
			if ((cluster->end & ~CLUSTER_MASK) == 0
			    || cache->nr_requests >= MAX_REQUEST_SOFT) {
				error = sync_cluster(cluster, 0);
			} else if (time_before(cluster->nextscan, jiffies)) {
				cluster->nextscan = jiffies + NFS_WRITEBACK_DELAY;
				error = scan_cluster(cluster, 0);
			}

		release:
			release_cluster(previous);
			previous = cluster;
			cluster  = cluster->hash_next;
			delay = 1 * HZ;
		}
		release_cluster(previous);

		if (error < 0) {
			delay = HZ >> 4;        /* Not enough memory */
			break;
		}
		hand = (hand + 1) & (CLUSTER_HASH_SIZE - 1);
	} while (hand != cache->hand);

	cache->hand = hand;

	dprintk("NFS: %4d clusterd back to sleep\n", task->tk_pid);
	task->tk_timeout = delay;
	rpc_sleep_on(&cluster_queue, task, NULL, NULL);
}

/*
 * Wait for the RPC backlog to clear.
 *
 * Interruptible by signals only if mounted with intr flag.
 */
int
nfs_wait_on_congest(struct inode *inode)
{
        struct rpc_clnt         *clnt = NFS_CLIENT(inode);
	struct rpc_xprt		*xprt = clnt->cl_xprt;
        struct wait_queue       wait = { current, NULL };
        sigset_t                oldmask;
        int retval, intr;

	if (!NFS_CONGESTED(inode))
		return 0;

	intr = NFS_SERVER(inode)->flags & NFS_MOUNT_INTR;
        rpc_clnt_sigmask(clnt, &oldmask);
	add_wait_queue(&xprt->cong_wait, &wait);

        for(;;) {
            	current->state = (intr) ? TASK_INTERRUPTIBLE
					: TASK_UNINTERRUPTIBLE;

                retval = 0;
                if (!NFS_CONGESTED(inode))
                        break;

                retval = -EINTR;
                if (intr && signalled())
                        break;

                schedule();
        }

	remove_wait_queue(&xprt->cong_wait, &wait);
	current->state = TASK_RUNNING;
	rpc_clnt_sigunmask(clnt, &oldmask);

	return retval;
}
