/*
 * linux/fs/nfs/write.c
 *
 * Writing file data over NFS.
 *
 * We do it like this: When a (user) process wishes to write data to an
 * NFS file, a write request is allocated that contains the RPC task data
 * plus some info on the page to be written, and added to the inode's
 * write chain. If the process writes past the end of the page, an async
 * RPC call to write the page is scheduled immediately; otherwise, the call
 * is delayed for a few seconds.
 *
 * Just like readahead, no async I/O is performed if wsize < PAGE_SIZE.
 *
 * Write requests are kept on the inode's writeback list. Each entry in
 * that list references the page (portion) to be written. When the
 * cache timeout has expired, the RPC task is woken up, and tries to
 * lock the page. As soon as it manages to do so, the request is moved
 * from the writeback list to the writelock list.
 *
 * Note: we must make sure never to confuse the inode passed in the
 * write_page request with the one in page->inode. As far as I understand
 * it, these are different when doing a swap-out.
 *
 * To understand everything that goes on here and in the NFS read code,
 * one should be aware that a page is locked in exactly one of the following
 * cases:
 *
 *  -	A write request is in progress.
 *  -	A user process is in generic_file_write/nfs_update_page
 *  -	A user process is in generic_file_read
 *
 * Also note that because of the way pages are invalidated in
 * nfs_revalidate_inode, the following assertions hold:
 *
 *  -	If a page is dirty, there will be no read requests (a page will
 *	not be re-read unless invalidated by nfs_revalidate_inode).
 *  -	If the page is not uptodate, there will be no pending write
 *	requests, and no process will be in nfs_update_page.
 *
 * FIXME: Interaction with the vmscan routines is not optimal yet.
 * Either vmscan must be made nfs-savvy, or we need a different page
 * reclaim concept that supports something like FS-independent
 * buffer_heads with a b_ops-> field.
 *
 * Copyright (C) 1996, 1997, Olaf Kirch <okir@monad.swb.de>
 *
 * Some parts
 * Copyright (C) 1999, Trond Myklebust <trond.myklebust@fys.uio.no>
 */

#include <linux/types.h>
#include <linux/malloc.h>
#include <linux/swap.h>
#include <linux/pagemap.h>

#include <linux/sunrpc/auth.h>
#include <linux/sunrpc/clnt.h>
#include <linux/nfs_fs.h>
#include <linux/nfs_cluster.h>
#include <asm/uaccess.h>

#define NFS_PARANOIA 1
#define NFSDBG_FACILITY		NFSDBG_PAGECACHE

/*
 * NFSv3 constants
 */
#define NFS_COMMIT_DELAY         (10 * HZ)

/*
 * This is the struct where the WRITE/COMMIT arguments go.
 */
struct nfs_write_data {
        struct nfs_cluster *    cluster;        /* cluster */
        unsigned short          index;          /* request index */
        unsigned short          count;          /* # of coalesced pages */
        struct nfs_writeargs    args;           /* argument struct */
        struct nfs_writeres     res;            /* result struct */
        struct nfs_fattr        fattr;
        unsigned int            commit;         /* commit id */
};

/*
 * Local function declarations
 */
static void                     nfs_writeback_done(struct rpc_task *);
static struct nfs_write_data *  nfs_write_rpcsetup(struct dentry*,
					 struct inode *inode,
                                        __u64 start, __u64 end, int flags);
static void                     nfs_writeback_attrs(struct inode *inode,
                                        struct nfs_fattr *fattr);
static int                      flush_wrequest(struct nfs_page *, int);



/* Hack for future NFS swap support */
#ifndef IS_SWAPFILE
# define IS_SWAPFILE(inode)	(0)
#endif

/*
 * Write a page synchronously.
 * Offset is the data offset within the page.
 */
static int
nfs_writepage_sync(struct dentry *dentry, struct inode *inode,
		   struct rpc_cred *cred, struct page *page,
		   unsigned long offset, unsigned int count)
{
	unsigned int	wsize = NFS_SERVER(inode)->wsize;
	int		result, refresh = 0, written = 0;
	u8		*buffer;
	struct nfs_fattr fattr;

	dprintk("NFS:      nfs_writepage_sync(%s/%s %d@%ld)\n",
		dentry->d_parent->d_name.name, dentry->d_name.name,
		count, page->offset + offset);

	buffer = (u8 *) page_address(page) + offset;
	offset += page->offset;

	do {
		if (count < wsize && !IS_SWAPFILE(inode))
			wsize = count;

		result = nfs_proc_write(NFS_DSERVER(dentry), NFS_FH(dentry),
					cred, IS_SWAPFILE(inode),
					offset, wsize, buffer, &fattr);

		if (result < 0) {
			/* Must mark the page invalid after I/O error */
			clear_bit(PG_uptodate, &page->flags);
			goto io_error;
		}
		if (result != wsize)
			printk("NFS: short write, wsize=%u, result=%d\n",
			wsize, result);
		refresh = 1;
		buffer  += wsize;
		offset  += wsize;
		written += wsize;
		count   -= wsize;
		/*
		 * If we've extended the file, update the inode
		 * now so we don't invalidate the cache.
		 */
		if (offset > inode->i_size)
			inode->i_size = offset;
	} while (count);

io_error:
	/* Note: we don't refresh if the call failed (fattr invalid) */
	if (refresh && result >= 0) {
		/* See comments in nfs_wback_result */
		/* N.B. I don't think this is right -- sync writes in order */
		if (fattr.size < inode->i_size)
			fattr.size = inode->i_size;
		if (fattr.mtime.seconds < inode->i_mtime)
			printk("nfs_writepage_sync: prior time??\n");
		/* Solaris 2.5 server seems to send garbled
		 * fattrs occasionally */
		if (inode->i_ino == fattr.fileid) {
			/*
			 * We expect the mtime value to change, and
			 * don't want to invalidate the caches.
			 */
			inode->i_mtime = fattr.mtime.seconds;
			nfs_refresh_inode(inode, &fattr);
		} 
		else
			printk("nfs_writepage_sync: inode %ld, got %u?\n",
				inode->i_ino, fattr.fileid);
	}

	nfs_unlock_page(page);
	return written? written : result;
}

/*
 * Write a page to the server. This will be used for NFS swapping only
 * (for now), and we currently do this synchronously only.
 */
int
nfs_writepage(struct file * filp, struct page *page)
{
	struct dentry	*dentry = filp->f_dentry;
	struct rpc_cred	*cred = NFS_FILE_CRED(filp);
	int		status;

	status = nfs_writepage_sync(dentry, dentry->d_inode, cred, page,
				    0, PAGE_SIZE);
	return status;
}

/*
 * Update and possibly write a cached page of an NFS file.
 *
 * XXX: Keep an eye on generic_file_read to make sure it doesn't do bad
 * things with a page scheduled for an RPC call (e.g. invalidate it).
 */
int
nfs_updatepage(struct file *file, struct page *page, const char *buf, unsigned long offset, unsigned int count, int sync)
{
	struct dentry	   *dentry = file->f_dentry;
	struct inode	   *inode = dentry->d_inode;
	struct nfs_page	   *req;
	struct nfs_cluster *cluster;
	int		    status;

	dprintk("NFS:      nfs_updatepage(%s/%s %d@%ld, sync=%d)\n",
		dentry->d_parent->d_name.name, dentry->d_name.name,
		count, page->offset+offset, sync);


	set_bit(PG_locked, &page->flags);
	/*
	 * If wsize is smaller than page size, update and write
	 * page synchronously.
	 */
	if (NFS_SERVER(inode)->wsize < PAGE_SIZE) {
		count -= copy_from_user((u8*)page_address(page) + offset,
					buf, count);
		if (!count) {
			status = -EFAULT;
			clear_bit(PG_uptodate, &page->flags);
			goto out_free;
		}
		status = nfs_writepage_sync(dentry, inode, NFS_FILE_CRED(file),
					    page, offset, count);
		return status;
	}

        /*
         * Try to find an NFS request corresponding to this page
         * and update it.
         * If there is one, we can be sure it is not yet being processed
         * because *we* hold the lock on the page.
         * If the request was not created by the calling process (or the
         * process was running under a different uid when creating it),
         * we must flush it out now. Lend it the page lock and perform
         * a synchronous RPC WRITE call.
         */
again:
        status = nfs_update_request(file, page, offset, count, &req, flush_wrequest);
	req->flags &= ~PG_UNLOCK_AFTER;
        if (status == 0) {
                status = req->cluster->flush(req, FLUSH_SYNC|FLUSH_WAIT);
                if (status >= 0)
                        goto again;
                /* XXX: when the server returned an error, check that it is
                 * propagated to the proper process (it should be).
                 */
        }
        if (status < 0) {
                /* Error when allocating cluster or request, or when
                 * copying used data.
                 */
                goto out_free;
        }

	count -= copy_from_user((u8*)page_address(page) + offset, buf, count);

	if (!count) {
		status = -EFAULT;
		clear_bit(PG_uptodate, &page->flags);
		goto out_free;
	}

	if (req->end >= req->start + PAGE_SIZE)
		set_bit(PG_uptodate, &page->flags);

        cluster = req->cluster;
        status = count;         /* unless we detect an error */

        /* If the user requested a sync write, do it now */
        if (sync) {
                int     error;

                error = flush_wrequest(req, FLUSH_SYNC|FLUSH_WAIT);
                if (error >= 0 && (error = file->f_error) < 0)
                        status = error;
		goto out_free;
        } else {
		cluster->count++;
		nfs_unlock_page(page);

               	/* If we wrote past the end of the page.
               	 * Call the strategy routine so it can send out a bunch
               	 * of requests.
               	 */
		 if (req->end >= req->start + PAGE_SIZE)
			nfs_strategy(cluster);

		release_cluster(cluster);
        }

        /* Cluster may have gone at this point */

done:
        dprintk("NFS:      nfs_updatepage returns %d (isize %ld)\n",
                                                status, inode->i_size);
        return status;
 out_free:
	nfs_unlock_page(page);
	goto done;

}

/*
 * Coalesce adjacent write requests
 */
static void
coalesce(struct nfs_page *req, struct nfs_write_data *data)
{
        struct nfs_cluster      *cluster = req->cluster;
        unsigned int            wpages, index, count, total;
        struct iovec            *iov;
        __u64                   pos, offset;

        /* Fixme: put this in the server struct one day */
        wpages = NFS_SERVER(cluster->file->f_dentry->d_inode)->wsize / PAGE_SIZE;
        if (wpages > NFS_WRITE_MAXIOV)
                wpages = NFS_WRITE_MAXIOV;

        /* Index of first request in this write call */
        pos         = req->end;
        index       = REQUEST_NR(req->start);
        data->index = index;
        data->count = 0;
        offset      = req->start & ~PAGE_MASK;
        total       = 0;

        goto first;

        while (index < CLUSTER_PAGES && data->count < wpages) {
                req = cluster->request[index];
                if (req == 0 || req->start != pos
                 || !IS_DIRTY(req) || !nfs_lock_page(req))
                        break;
        first:
                count = req->end - req->start;
                iov = data->args.iov + data->count;
                iov->iov_base = (u8 *) page_address(req->page) + offset;
                iov->iov_len = count;
                total += count;
                data->count++;
                index++;

                pos = req->end;
                offset = 0;

        }

        dprintk("NFS: coalesce gathered %d pages (first %d)\n",
                                data->count, data->index);
        data->args.nriov = data->count;
        data->args.count = total;
}

/*
 * Create an RPC task for the given write request and kick it.
 * The page must have been locked by the caller.
 *
 * It may happen that the page we're passed is not marked dirty.
 * This is the case if nfs_updatepage detects a conflicting request
 * that has been written but not committed.
 *
 * This routine should always succeed when given FLUSH_SYNC.
 */
static int
flush_wrequest(struct nfs_page *req, int how)
{
	struct nfs_write_data	*data;
	struct nfs_cluster	*cluster = req->cluster;
	struct file		*filp = cluster->file;
	struct dentry		*dentry = filp->f_dentry;
	struct inode		*inode = dentry->d_inode;
	struct rpc_clnt		*clnt = NFS_CLIENT(inode);
	struct rpc_task		*task;
	int			flags = 0;
	u32			proc;
	sigset_t		oldset;

	cluster->count++;

	if (!IS_DIRTY(req)) {
		if (!(how & FLUSH_STABLE)) {
			printk("NFS: flush_wrequest called for clean page!\n");
			how |= FLUSH_STABLE;
		}

		/* Make sure we mark the page dirty so that our book-
		 * keeping is consistent */
		nfs_mark_request_dirty(cluster, req);
	}

	/* Set the initial flags for the task. PRIORITY means that
	 * the RPC code will set need_resched whenever the task becomes
	 * runnable. ASYNC means that the task will be asynchronously.
	 * In addition to that, this means that rpc_allocate will not
	 * block when it can't get any memory. Without ASYNC, rpc_new_task
	 * may block while allocating memory.
	 * That's bad for swapping, but good for general sanity.
	 *
	 * (We also set ASYNC later when we see that the caller doesn't
	 * want to wait for the task's completion).
	 */
	flags = (how & FLUSH_SYNC)? RPC_TASK_PRIORITY : RPC_TASK_ASYNC;

	rpc_clnt_sigmask(clnt, &oldset);

	task = rpc_new_task(clnt, nfs_writeback_done, flags);
	data = nfs_write_rpcsetup(dentry, inode, req->start, req->end, flags);

	if (task == NULL || data == NULL) {
		/* We failed to allocate the task. To avoid any hassle
		 * in the upper layers, release the page lock if we own it. */
		if (req->flags & PG_UNLOCK_AFTER) {
			req->flags &= ~PG_UNLOCK_AFTER;
			nfs_unlock_page(req->page);
		}
		if (task)
			rpc_release_task(task);
		release_cluster(cluster);
		rpc_clnt_sigunmask(clnt, &oldset);
		return -ENOMEM;
	}
	req->task = task;

	/* Unless the caller wants to wait on the request, make the RPC
	 * task async now. */
	if (!(how & FLUSH_WAIT))
		task->tk_flags |= RPC_TASK_ASYNC;

	data->cluster = cluster;

	/* Put one or more pages of data into call */
	coalesce(req, data);
	cluster->pending += data->count;

	/* Hack for NFSv2 as it never returns the size written */
	data->res.count = data->args.count;

	/* Finalize the task. */
	task->tk_calldata = data;

	task->tk_cred = NFS_FILE_CRED(filp);
	rpcauth_holdcred(task);

	proc = NFSPROC_WRITE;

	dprintk("NFS: %4d initiated write call (req %s/%s count %d nriov %d)\n",
		task->tk_pid,
		dentry->d_parent->d_name.name, dentry->d_name.name,
		data->args.count, data->args.nriov);
	rpc_call_setup(task, proc, &data->args, &data->res, 0);
	rpc_execute(task);

	rpc_clnt_sigunmask(clnt, &oldset);
	return 0;
}


/*
 * Allocate the argument/result storage required for the RPC call.
 */
static struct nfs_write_data *
nfs_write_rpcsetup(struct dentry * dentry, struct inode *inode,
		   __u64 start, __u64 end, int flags)
{
        struct nfs_write_data   *data;
        void                    *ptr;

        if (!(ptr = rpc_allocate(flags, sizeof(*data))))
                return NULL;

        /* Set up the RPC argument and reply structs */
        data = (struct nfs_write_data *) ptr;
        data->args.fh     = NFS_FH(dentry);
        data->res.fattr   = &data->fattr;
        data->res.count   = end - start;
        data->args.offset = start;
        data->args.count  = end - start;
        return data;
}


/*
 * This function is called when the WRITE call is complete.
 */
static void
nfs_writeback_done(struct rpc_task *task)
{
	struct nfs_cluster      *cluster;
	struct nfs_write_data   *data;
	struct nfs_writeargs    *argp;
	struct nfs_writeres     *resp;
	struct nfs_page         *req;
	struct inode            *inode;
	unsigned int            nr, count;

	dprintk("NFS: %4d nfs_writeback_done (status %d)\n",
		task->tk_pid, task->tk_status);

	if (!(data = (struct nfs_write_data *) task->tk_calldata)) {
		printk("NFS: no write data for RPC task?!\n");
		return;
	}

	cluster = data->cluster;
	argp    = &data->args;
	resp    = &data->res;
	inode   = cluster->file->f_dentry->d_inode;

	/* Update attributes as result of writeback. */
	nfs_writeback_attrs(inode, resp->fattr);

#ifdef RPC_DEBUG
	for (nr = data->index, count = data->count; count--; nr++) {
		if ((req = cluster->request[nr]) != NULL)
			continue;
		printk("NFS: request disappeared in write callback.\n");
		task->tk_status = -EIO;
		break;
	}
#endif
	/* We can't handle that yet but we check for it nevertheless */
	if (resp->count < argp->count && task->tk_status >= 0) {
		static unsigned long    complain = 0;
		if (time_before(complain, jiffies)) {
			printk("NFS: Server wrote less than requested.\n");
			complain = jiffies + 300 * HZ;
		}
		/* Can't do anything about it right now except throw
		 * an error. */
		task->tk_status = -EIO;
	}


	if (task->tk_status < 0) {
		/* A WRITE error occurred. Report the error back to the
		 * application. */
		cluster->file->f_error = task->tk_status;
		for (nr = data->index, count = data->count; count--; nr++) {
			if ((req = cluster->request[nr]) != NULL)
				clear_bit(PG_uptodate, &req->page->flags);
		}
	}

	/* Release requests */
	for (nr = data->index, count = data->count; count--; nr++) {
		if (!(req = cluster->request[nr]))
			continue;
		req->task = NULL;
		nfs_release_request(req);
	}

	/* Cluster book-keeping. */
	cluster->dirty -= data->count;

	/*
	 * If all write requests for that cluster have completed, commit.
	 */
	cluster->pending -= data->count;

	kfree(task->tk_calldata);

	if (!cluster->pending)
		wake_up(&cluster->wait);
	release_cluster(cluster);
}

/*
 * Handle the file attributes returned in a WRITE response.
 * Beware: Normally, the reply to an async write requests arrives
 * after the user process has written a lot more data to the
 * cache, hence the file size returned by the server is completely
 * bogus. We therefore fake the returned file size.
 *
 * Also, the pre-op inode size returned by the server reflects the
 * current length of the file on the server, not our local copy. Hence
 * NFSv3 will always flush the caches, potentially sending off
 * huge amounts of data to stable storage. Therefore, we also fake
 * the returned pre-op size.
 *
 * The inode number check tries to avoid a possible
 * Solaris 2.4/2.5 server bug.
 */
static void
nfs_writeback_attrs(struct inode *inode, struct nfs_fattr *fattr)
{

	if (inode->i_ino != fattr->fileid)
		return;
	if (fattr->mtime.seconds < inode->i_mtime)
		return;
        if (fattr->size < inode->i_size)
                fattr->size = inode->i_size;

	inode->i_size  = fattr->size;
	inode->i_mtime = fattr->mtime.seconds;

        /* Now refresh the inode */
        nfs_refresh_inode(inode, fattr);
}

